/*
 * guess.c
 *
 * All rights reserved. Copyright (C) 1994,1998 by NARITA Tomio
 */

#include <import.h>
#include <decode.h>
#include <big5.h>
#include <utf.h>
#ifndef MSDOS /* IF NOT DEFINED */
#include <unimap.h>
#include <unirev.h>
#endif /* MSDOS */
#include <begin.h>
#include <guess.h>

public char GuessCodingSystem( char *str, int length, char defaultEuc )
{
  int i;
  char ch;
  int big5Penalty = 0;
  int sjisPenalty = 0;
  int utf8Penalty = 0;
  int eucjapanPenalty = 0;
  int eucPenalty = 0;

  /*
   * Locale specific implementation is required.
   */
  for( i = 0 ; i < length ; i++ ){
    /*
     * check for simple euc
     */
    ch = str[ i ];
    if( 0x80 & ch ){
      if( SS2 == ch || SS3 == ch ){
	eucPenalty++;
	continue;
      }
      if( !IsEucByte( ch ) )
	eucPenalty++;
      if( ++i >= length )
	break;
      ch = str[ i ];
      if( !IsEucByte( ch ) )
	eucPenalty++;
    }
  }
  for( i = 0 ; i < length ; i++ ){
    /*
     * check for euc-japan or euc-taiwan
     */
    ch = str[ i ];
    if( 0x80 & ch ){
      if( SS2 == ch ){
	if( ++i >= length )
	  break;
	ch = str[ i ];
	if( !IsKatakanaByte( ch ) )
	  eucjapanPenalty++;
	continue;
      }
      if( SS3 == ch ){
	if( ++i >= length )
	  break;
	ch = str[ i ];
      }
      if( !IsEucByte( ch ) )
	eucjapanPenalty++;
      if( ++i >= length )
	break;
      ch = str[ i ];
      if( !IsEucByte( ch ) )
	eucjapanPenalty++;
    }
  }
  for( i = 0 ; i < length ; i++ ){
    /*
     * check for big5
     */
    ch = str[ i ];
    if( 0x80 & ch ){
      if( !IsBig5Byte1( ch ) )
	big5Penalty++;
      ch = str[ ++i ];
      if( i >= length )
	break;
      if( !IsBig5Byte2( ch ) )
	big5Penalty++;
    }
  }
  for( i = 0 ; i < length ; i++ ){
    /*
     * check for shift-jis
     */
    ch = str[ i ];
    if( 0x80 & ch ){
      if( IsKatakanaByte( ch ) )
	continue;
      if( !IsShiftJisByte1( ch ) )
	sjisPenalty++;
      ch = str[ ++i ];
      if( i >= length )
	break;
      if( !IsShiftJisByte2( ch ) )
	sjisPenalty++;
    }
  }
#ifndef MSDOS /* IF NOT DEFINED */
  { int idx;
    decoding_penalty = 0;
    idx = length;
    IstrFree( Decode( UTF_8, str, &idx ) );
    utf8Penalty = decoding_penalty;
  }
#endif /* MSDOS */

  if( 0 == eucPenalty )
    return defaultEuc;
  else if( 0 == eucjapanPenalty )
    return EUC_JAPAN;
  else if( 0 == big5Penalty )
    return BIG_FIVE;
  else if( 0 == sjisPenalty )
    return SHIFT_JIS;
#ifndef MSDOS /* IF NOT DEFINED */
  else if( 0 == utf8Penalty )
    return UTF_8;
#endif /* MSDOS */

  return ISO_8859_1;
}

public char GuessHz( char *str, int length )
{
  boolean_t gb = FALSE;
  int i, hzPenalty = 0;
  char ch;

  for( i = 0 ; i < length ; i++ ){
    /*
     * check for HZ
     */
    ch = str[ i ];
    if( '~' == ch ){
      if( ++i >= length )
	break;
      ch = str[ i ];
      if( FALSE == gb ){
	if( '{' == ch ){
	  gb = TRUE;
	} else if( '~' == ch || '}' == ch ){
	  /* do nothing */
	} else
	  hzPenalty++;
      } else {
	if( '}' == ch )
	  gb = FALSE;
      }
    }
  }

  return 0 == hzPenalty;
}


public void AdjustPatternCharset( char inputCodingSystem,
				 char keyboardCodingSystem,
				 char defaultCodingSystem,
				 i_str_t *istr )
{
  if( FALSE == adjust_charset )
    return;

#ifndef MSDOS /* IF NOT DEFINED */
  if( IsUtfEncoding( inputCodingSystem ) || IsUtfEncoding( keyboardCodingSystem ) ){
    if( IsUtfEncoding( inputCodingSystem ) && !IsUtfEncoding( keyboardCodingSystem ) )
      ConvertToUNI( istr );
    else if( !IsUtfEncoding( inputCodingSystem ) && IsUtfEncoding( keyboardCodingSystem ) ){
      if( AUTOSELECT == inputCodingSystem )
	inputCodingSystem = defaultCodingSystem;
      ConvertFromUNI( istr, inputCodingSystem );
    }
  } else
#endif /* MSDOS */
    if( BIG_FIVE == inputCodingSystem && BIG_FIVE != keyboardCodingSystem ){
      ConvertCNStoBIG5( istr );
    } else if( BIG_FIVE != inputCodingSystem && BIG_FIVE == keyboardCodingSystem ){
      ConvertBIG5toCNS( istr );
    }
}
